#include <cppunit/extensions/HelperMacros.h>

#include <vector>
#include <iomanip>

#include <boost/foreach.hpp>
#include <boost/format.hpp>
#include <boost/make_shared.hpp>
#include <boost/shared_ptr.hpp>

#include "src/hwdesc.h"
#include "src/hwdesc_attributes.h"
#include "src/charging_pulse.h"
#include "src/util.h"

//
// The terminal blocks must not charge their internal capacities in the same
// short interval of the phase. We should use each of the 32 charge slots
// equally
//

typedef boost::shared_ptr<const dSA::Device> const_device_p;

using namespace dSA;

class PulseConfigTest : public CppUnit::TestFixture
{
    CPPUNIT_TEST_SUITE(PulseConfigTest);
    CPPUNIT_TEST(testSortByTemp);
    CPPUNIT_TEST(testPrintMoreDevicesThanSlots);
    CPPUNIT_TEST(testPrintMoreSlotsThanDevices);
    CPPUNIT_TEST(testNonMainDevices);
    CPPUNIT_TEST_SUITE_END();
public:
    void setUp();
    void tearDown();
    void testSortByTemp();
    void testPrintMoreDevicesThanSlots();
    void testPrintMoreSlotsThanDevices();
    void testNonMainDevices();
};

CPPUNIT_TEST_SUITE_REGISTRATION(PulseConfigTest);

static uint16_t randTempAttr()
{
    return rand() % 150;
}

static boost::shared_ptr<Device> createDevice(int temperature)
{
    boost::shared_ptr<Device> dev =
        boost::make_shared<Device>(DSUID_NULL, DSUID_BROADCAST);
    dev->setTemperature(temperature);
    return dev;
}

devices_t createDeviceList(int count)
{
    devices_t devs;
    for (int i = 0; i < count; i++)
    {
        int temp = attributeToTemperature(randTempAttr());
        devs.push_back(createDevice(temp));
    }
    return devs;
}

void PulseConfigTest::setUp()
{
    // using random data, gives better coverage
    // if unit-test fails, grab the 'srand init:' from
    // test-suite.log output, init random generator
    // with that value to reproduce error

    time_t init = time(0);
    std::cout << "srand init: " << init << "\n";
    srand(init);

    //srand(1433757907); //< std::sort crash
}

void PulseConfigTest::tearDown()
{
}

void PulseConfigTest::testSortByTemp()
{
    devices_t devs = createDeviceList(77);

    // sort needs random access iterator
    std::vector<const_device_p> t(devs.begin(), devs.end());
    std::sort(t.begin(), t.end(), sortByTemperature);

    int temp = MIN_TEMP;
    BOOST_FOREACH(const_device_p elt, t)
    {
        CPPUNIT_ASSERT(elt->getLastKnownTemperature() >= temp);
        temp = elt->getLastKnownTemperature();
    }
}

std::vector<int> createHistogram(cp_assignments_t assigns)
{
    std::vector<int> hist(charge_slot_count, 0);

    BOOST_FOREACH(cp_assignment_t elt, assigns)
    {
        hist[elt.second]++;
    }
    CPPUNIT_ASSERT(hist.size() <= charge_slot_count);
    return hist;
}

void plotHistogram(std::vector<int> hist)
{
    int max = *std::max_element(hist.begin(), hist.end());

    std::cout << std::endl;
    for (int line = max; line > 0; line--)
    {
        std::cout << std::setw(3) << line << " dev : |";
        BOOST_FOREACH(int elt, hist)
        {
            if (elt >= line)
            {
                std::cout << '+';
            }
            else
            {
                std::cout << ' ';
            }
        }
        std::cout << "|\n";
    }

    std::cout << "          +----------------------------------------+\n";
    std::cout << "slot id :  0123456789012345678901234567890123456789\n";
    std::cout << "           0         1         2         3\n";
}


void dumpAssignments(cp_assignments_t assigns)
{
    std::cout << "\n-- slot allocations --\n";
    BOOST_FOREACH(cp_assignment_t elt, assigns)
    {
        std::cout << std::fixed << std::setprecision(1);
        std::cout << "temp: " << std::setw(5)
            << (elt.first->getLastKnownTemperature() / 10.0) << " C "
            << "slot:" << std::setw(3) << elt.second << "\n";
    }
}

void PulseConfigTest::testPrintMoreDevicesThanSlots()
{
    // create about 3 times the numbers of slot
    size_t ct = 2 * charge_slot_count + charge_slot_count / 3;

    devices_t devs = createDeviceList(ct);
    CPPUNIT_ASSERT_EQUAL(devs.size(), ct);

    std::cout
        <<  "devices:" << devs.size()
        << " slots:" << charge_slot_count
        << std::endl;

    cp_assignments_t result = assignChargingPulse(devs);

    std::vector<int> hist = createHistogram(result);
    plotHistogram(hist);
    // dumpAssignments(result);

    std::cout << std::endl;
}

void PulseConfigTest::testPrintMoreSlotsThanDevices()
{
    // create a scarce list
    devices_t devs = createDeviceList(charge_slot_count / 5);

    std::cout << std::endl;
    std::cout << "devices: " << devs.size()
        << " slots:" << charge_slot_count
        << std::endl;

    cp_assignments_t result = assignChargingPulse(devs);

    std::vector<int> hist = createHistogram(result);
    plotHistogram(hist);
    dumpAssignments(result);

    std::cout << std::endl;
}

static void setSlaveDevice(Device &dev)
{
    // see getHardwareInfoString: sw-sds201
    unsigned color = 0x7; // sw
    unsigned devtype = 0x6; // sds

    unsigned function_id = (color + 1) << 12;
    unsigned product_id = devtype << 10 | 201;

    dev.setFunctionId(function_id);
    dev.setProductId(product_id);
}

void PulseConfigTest::testNonMainDevices()
{
    device_p dev = createDevice(209);
    CPPUNIT_ASSERT(dev->isMainDevice());

    setSlaveDevice(*dev);
    CPPUNIT_ASSERT(!dev->isMainDevice());

    devices_t non_main;
    devices_t devs = createDeviceList(17);
    BOOST_FOREACH(devices_t::value_type elt, devs)
    {
        non_main.push_back(elt);
        setSlaveDevice(*elt);
        if (non_main.size() >= 7)
        {
            break;
        }
    }

    cp_assignments_t result = assignChargingPulse(devs);
    CPPUNIT_ASSERT_EQUAL(devs.size() - non_main.size(), result.size());
}
